import os
import numpy
import datetime
import netCDF4
import csv
import re
import textwrap
from .spacel import Field, FieldList, Space
from .data   import Data, set_Data_attributes
import time


#_read_data = False
_word_size = 4
_pp_rmdi   = -1e+30 

try:
    _UMVERSION = os.environ['UMVERSION']
    _UMVERSION = _UMVERSION.replace('.', '0', 1)
except KeyError:
    pass

_t = {}
_t['space_attribute']=0.0
_t['pp_i']=0.0
_t['pp_h']=0.0
_t['pp_h_io']=0.0
_t['pp_h_nct']=0.0
_t['pp_d']=0.0
_t['dim']=0.0
_t['aux']=0.0
_xrange45 = range(45)
_xrange19 = range(19)
_number_regex = '([-+]?\d*\.?\d+(e[-+]?\d+)?)'
_cache = {'runid' : {},
          'latlon': {},
          'time'  : {},
          }
_lhd_items = ('lbyr', 'lbmon', 'lbdat', 'lbhr', 'lbmin', 'lbday',
             'lbyrd', 'lbmond', 'lbdatd', 'lbhrd', 'lbmind',
             'lbdayd', 'lbtim', 'lbft', 'lblrec', 'lbcode', 'lbhem',
             'lbrow', 'lbnpt', 'lbext', 'lbpack', 'lbrel', 'lbfc',
             'lbcfc', 'lbproc', 'lbvc', 'lbrvc', 'lbexp', 'lbegin', 
             'lbnrec', 'lbproj', 'lbtyp', 'lblev', 'lbrsvd1',
             'lbrsvd2', 'lbrsvd3', 'lbrsvd4', 'lbsrce', 'lbuser1',
             'lbuser2', 'lbuser3', 'lbuser4', 'lbuser5', 'lbuser6',
             'lbuser7')

(lbyr, lbmon, lbdat, lbhr, lbmin, lbday,
 lbyrd, lbmond, lbdatd, lbhrd, lbmind,
 lbdayd, lbtim, lbft, lblrec, lbcode, lbhem,
 lbrow, lbnpt, lbext, lbpack, lbrel, lbfc,
 lbcfc, lbproc, lbvc, lbrvc, lbexp, lbegin, 
 lbnrec, lbproj, lbtyp, lblev, lbrsvd1,
 lbrsvd2, lbrsvd3, lbrsvd4, lbsrce, lbuser1,
 lbuser2, lbuser3, lbuser4, lbuser5, lbuser6,
 lbuser7) = range(45)

(brsvd1, brsvd2, brsvd3, brscv4, 
 bdatum, bacc, blev, brlev, bhlev, bhrlev,
 bplat, bplon, bgor,
 bzy, bdy, bzx, bdx, bmdi, bmks) = range(19)



# Assign integer positions 0-18 to real header item names
_bhd_items = ('brsvd1', 'brsvd2', 'brsvd3', 'brscv4', 
              'bdatum', 'bacc', 'blev', 'brlev', 'bhlev', 'bhrlev',
              'bplat', 'bplon', 'bgor',
              'bzy', 'bdy', 'bzx', 'bdx', 'bmdi', 'bmks')

_extra_data_name = {
    1  : 'x',
    2  : 'y',
    3  : 'y_domain_lower_bound',
    4  : 'x_domain_lower_bound',
    5  : 'y_domain_upper_bound',
    6  : 'x_domain_upper_bound',
    7  : 'z_domain_lower_bound',
    8  : 'x_domain_upper_bound',
    10 : 'title',
    11 : 'domain_title',
    12 : 'x_lower_bound',
    13 : 'x_upper_bound',
    14 : 'y_lower_bound',
    15 : 'y_upper_bound',
    }

# Tranlate axis codes to CF 'atandard_name' values
_coord_standard_name = {
    1   : 'air_pressure', 
    2   : 'altitude',
    3   : 'atmosphere_hybrid_sigma_pressure_coordinate',
    4   : 'depth',
    5   : 'model_level_number',        
    6   : 'air_potential_temperature',
    7   : 'atmosphere_sigma_coordinate',
    10  : 'latitude',
    11  : 'longitude',
    13  : 'region',
    14  : 'atmosphere_hybrid_height_coordinate',
    15  : 'height',
    20  : 'time',  # (gregorian)
    23  : 'time',  # (360_day)
    40  : None,    # pseudolevel
    }

# Tranlate axis codes to CF 'long_name' values
_coord_long_name = {
    40 : 'pseudolevel',
    }

# Tranlate axis codes to CF 'units' values
_coord_units = {
    1   : 'hPa',           # air_pressure                      
    2   : 'm',             # altitude         
    3   : '1',             # atmosphere_hybrid_sigma_pressure_coordinate
    4   : 'm',             # depth                                  
    5   : '1',             # model_level_number                         
    6   : 'K',             # air_potential_temperature
    7   : '1',             # atmosphere_sigma_coordinate               
    10  : 'degrees_north', # latitude                               
    11  : 'degrees_east',  # longitude                                
    13  : 'string',        # region                                     
    14  : '1',             # atmosphere_hybrid_height_coordinate          
    15  : 'm',             # height                                      
    20  : 'days',          # time (gregorian)                    
    23  : 'days',          # time (360_day)
    40  : '1',             # pseudolevel
    }

# Tranlate axis codes to CF 'axis' values
_coord_axis = {
    1   : 'Z',   # air_pressure                       
    2   : 'Z',   # altitude                                     
    3   : 'Z',   # atmosphere_hybrid_sigma_pressure_coordinate  
    4   : 'Z',   # depth                                        
    5   : 'Z',   # model_level_number                          
    6   : 'Z',   # air_potential_temperature
    7   : 'Z',   # atmosphere_sigma_coordinate                
    10  : 'Y',   # latitude                                     
    11  : 'X',   # longitude                                    
    13  : None,  # region                                       
    14  : 'Z',   # atmosphere_hybrid_height_coordinate          
    15  : 'Z',   # height                                       
    20  : 'T',   # time (gregorian)                                         
    23  : 'T',   # time (360_day)                                         
    40  : None,  # pseudolevel                                    
    }

# Tranlate axis codes to CF 'positive' alues
_coord_positive = {
    1   : 'down',  # air_pressure                     
    2   : 'up',    # altitude                                  
    3   : 'down',  # atmosphere_hybrid_sigma_pressure_coordinate 
    4   : 'down',  # depth                                     
    5   : None,    # model_level_number                         
    6   : 'up',    # air_potential_temperature
    7   : 'down',  # atmosphere_sigma_coordinate               
    10  : None,    # latitude                                   
    11  : None,    # longitude                                   
    14  : 'up',    # region                                     
    13  : None,    # atmosphere_hybrid_height_coordinate         
    15  : 'up',    # height                                      
    20  : None,    # time (gregorian)                                          
    23  : None,    # time (360_day)                                        
    40  : None,    # pseudolevel                                    
    }

# Translate LBVC codes to axis codes
_lbvc_codes = {    
    9   :  3,  # atmosphere_hybrid_sigma_pressure_coordinate
    65  : 14,  # atmosphere_hybrid_height_coordinate 
    8   :  1,  # air_pressure             
    6   :  5,  # model_level_number          
    1   :  2,  # altitude  
    10  :  7,  # atmosphere_sigma_coordinate   ## dch check
    19  :  6,  # air_potential_temperature
    176 : 10,  # latitude    
    177 : 11,  # longitude      
    }

# Set up a tuple of characters used in decoding runids
_characters = ('a','b','c','d','e','f','g','h','i','j',
               'k','l','m','n','o','p','q','r',
               's','t','u','v','w','x','y','z',
               '0','1','2','3','4','5','6','7','8','9')
   

class Variable(object):
    ''' 
    
    '''

    __slots__ = ['_file',
                 '_start',
                 '_shape',
                 '_dtype',
                 '_FillValue',
                 'scale_factor',
                 'add_offset',
                 ]


    def __init__(self, **kwargs):
        _file = None
        [setattr(self, attr, kwargs[attr]) for attr in kwargs]
    #--- End: def

    def __getitem__(self, index):

        array = self.numpy_array()
        return array[index]
    #--- End: def

    @property
    def shape(self):
        '''

Data's dimension sizes.

:returns: A tuple

    '''
        return self._shape
    #--- End: def

    @property
    def size(self):
        '''

Number of elements in the data.

:returns: A non-negative integer
    '''
        return numpy.prod(self.shape)
    #--- End: def

    @property
    def ndim(self):
        '''

Number of data dimensions.

:returns: A non-negative integer
    '''
        return len(self.shape)
    #--- End: def

    @property
    def dtype(self):
        '''

Data-type of the data elements.

:returns: A numpy dtype object.

    '''
        return self._dtype
    #--- End: def

    def close(self):
        self._file.close()
    #--- End: def

    def numpy_array(self):
        ''' Read data from disk into a numpy array'''

        # Make sure that the file is open
        if self._file.closed:
            self._file = open(self._file.name, 'rb')

        # Jump to the correct place in the file
        self._file.seek(self._start, os.SEEK_SET)

        # Read the data array
        array = numpy.fromfile(self._file,
                               dtype=self.dtype,
                               count=self.size).reshape(self.shape)

        # Convert to a masked array, if _FillValue is set and the
        # array contains missing data
        try:
            _FillValue = self._FillValue
        except AttributeError:
            pass
        else:
            mask = numpy.zeros(array.shape, dtype=numpy.bool)
            mask += array == _FillValue # masked_values?
            array = numpy.ma.array(array, mask=mask, fill_value=_FillValue)

        # Unpack the array using scale_factor and add_offset, if
        # available
        try:
            if self.scale_factor != 1.0:
                array *= self.scale_factor
        except AttributeError:
            pass
        
        try:
            if self.add_offset != 0.0:
                array += self.add_offset
        except AttributeError:
            pass
        
        # Return the array
        return data
    #--- End: def

    def equals(self, other, rtol=None, atol=None):

        # Compare variables which are the same
        if id(self) == id(other):
            return True
    
        # Compare object which are both of the same type
        if type(self) == type(other):

            # Check that each instance has the same relevant property
            # values
            if (self._file != other._file and not
                os.path.samefile(self._file.name, other._file.name)):
                return False

            for prop in ('_start', '_dtype', '_shape'):
                x = getattr(self, prop)
                y = getattr(other, prop)
                if x != y:
                    return False                

            # Still here? Then they're the same
            return True
    
        if isinstance(other, numpy.ndarray):
            if rtol is None:
                rtol = cf.RTOL
            if atol is None:
                atol = cf.ATOL        
            print 'ho hum de hum'
            return cf.equals(self.numpy_array, other, rtol=rtol, atol=atol)

    #--- End: def

#--- End: class


class VariableCalc(Variable):

    __slots__ = []

    def numpy_array(self):
        '''

        :return: a numpy array '''
        array = self._pseudo_origin + \
            numpy.arange(1,self.size+1, dtype=self.dtype)*self._delta

        return array
    #--- End: def


    def equals(self, other, rtol=None, atol=None):

        if rtol is None:
            rtol = cf.RTOL
        if atol is None:
            atol = cf.ATOL                     

        if isinstance(other, numpy.ndarray):
            print 'pom, pom, pom'
            return cf.equals(self.numpy_array, other, rtol=rtol, atol=atol)

        # Check that each instance has the same relevant property
        # values
        for prop in ('_pseudo_origin', '_delta'):
            try:
                x = getattr(self, prop)
                y = getattr(other, prop)
                if not abs(x-y) <= atol + rtol*abs(y):
                    return False                
            except AttributeError:
                return False        

        try:
            x = self._shape
            y = other._shape
            if x != y:
                return False                
        except AttributeError:
            return False
        
        # Still here? Then they're the same
        return True

#        # Compare variables which are the same
#        if id(self) == id(other):
#            return True
    #--- End: def

#--- End: class


class VariableCalcBounds(VariableCalc):

    __slots__ = []

    def numpy_array(self):

        # return coordinate bound values
        array      = numpy.empty(self.shape, dtype=self.dtype)

        array[:,0] = (self._pseudo_origin + 0.5*self._delta + 
                      numpy.arange(self.shape[0], dtype=self.dtype)*self._delta)

        array[:,1] = array[:,0] + self._delta
    
        return array
    #--- End: def

#--- End: class


class VariableBounds(Variable):

    '''  '''

    __slots__ = ['_lower',
                 '_upper',
                 ]

    def __init__(self, lower, upper):
        self._lower = lower
        self._upper = upper

        self._dtype = self._lower.dtype
        self._shape = self._lower.shape + (2,)        
    #--- End: def
        
    def numpy_array(self):
        try:
            lower = self._lower.numpy_array()
        except AttributeError:
            lower = self._lower

        try:
            upper = self._upper.numpy_array()
        except AttributeError:
            upper = self._upper

        return numpy.column_stack((lower, upper))
    #--- End: def
#--- End: class

class PPMetadata(object):

    def __init__(self):
        '''
Set attributes giving the default data type for reals and
integers. The basic assumption is that data are native endian.
'''
        self.float32      = numpy.dtype('<f4').newbyteorder('=')
        self.int32        = numpy.dtype('<i4').newbyteorder('=')
        self.str32        = numpy.dtype('<S4').newbyteorder('=')
        self.file_float32 = float32
        self.file_int32   = int32
        self.file_str32   = str32
    #--- End: def

    @property
    def printfdr(self):
        '''
        '''
        fdr = []
        for i in range(45):
            fdr.append('%s:_%s' % (_lhd_items[i].upper(), self.lhd[i]))
        for i in range(19):
            fdr.append('%s:_%s' % (_bhd_items[i].upper(), self.bhd[i]))
        fdr = textwrap.fill( ' '.join(fdr), width=79)
        fdr = [fdr.replace(':_', ': ')]

        if self.extra:
            fdr.append('EXTRA DATA:')
            for key in self.extra:
                fdr.append('%s: %s' % (key, repr(self.extra[key])))
        
        print '\n'.join(fdr)
    #--- End: def

#--- End: class
def unrotated_latlon(rotated_lat, rotated_lon, pole_lat, pole_lon):

    PI          = numpy.pi
    PI_OVER_180 = PI/180.0

    # Make sure pole_lon is between 0 and 360 and then convert it to
    # radians
    pole_lon = pole_lon % 360.0
    pole_lon *= PI_OVER_180
    pole_lat *= PI_OVER_180

    cos_pole_lat = numpy.cos(pole_lat)
    sin_pole_lat = numpy.sin(pole_lat)

    # Create appropiate copies of the input rotated arrays
    rot_lon = rotated_lon.copy()
    rot_lat = rotated_lat.view()

    # Make sure rotated longitudes are between -180 and 180
    while numpy.max(rot_lon) >= 180.0:
        rot_lon = numpy.where(rot_lon >= 180.0, rot_lon-360, rot_lon)
    while numpy.min(rot_lon) < -180.0:
        rot_lon = numpy.where(rot_lon < -180.0, rot_lon+360, rot_lon)

    # Create 2d arrays of rotated latitudes and longitudes in radians
    nlat = rot_lat.size
    nlon = rot_lon.size
    rot_lon = numpy.resize(numpy.deg2rad(rot_lon), (nlat, nlon))    
    rot_lat = numpy.resize(numpy.deg2rad(rot_lat), (nlon, nlat))
    rot_lat = rot_lat.reshape((nlat, nlon))

    # Find unrotated latitudes
    CPART = numpy.cos(rot_lon) * numpy.cos(rot_lat)
    sin_rot_lat = numpy.sin(rot_lat)
    x = cos_pole_lat * CPART + sin_pole_lat * sin_rot_lat
    x = numpy.clip(x, -1.0, 1.0)
    unrotated_lat = numpy.arcsin(x)
    
    # Find unrotated longitudes
    x = -cos_pole_lat*sin_rot_lat + sin_pole_lat*CPART
    x /= numpy.cos(unrotated_lat)   # dch /0 or overflow here? surely
                                    # lat could be ~+-pi/2? if so,
                                    # does x ~ cos(lat)?
    x = numpy.clip(x, -1.0, 1.0)
    unrotated_lon = -numpy.arccos(x)
    unrotated_lon = numpy.where(rot_lon > 0.0, unrotated_lon, -unrotated_lon)
    SOCK = pole_lon
    if abs(SOCK) > cf.ATOL:
        SOCK = 0
    else:
        SOCK = pole_lon - PI            
    unrotated_lon += SOCK

    # Convert unrotated latitudes and longitudes to degrees
    unrotated_lat = numpy.rad2deg(unrotated_lat)
    unrotated_lon = numpy.rad2deg(unrotated_lon)

    # Return unrotated latitudes and longitudes
    return numpy.rollaxis(numpy.dstack((unrotated_lat, unrotated_lon)), 2)
#--- End: def


def _extra_data(ppfile, m=None):
    '''
    Excerpt from UMDP F3 vn7.8
    --------------------------
    The data record may include 'extra data' in addition to the usual
    grid-point values. Cross-section fields must always be followed by
    extra data giving the x- and y-coordinate values of each column
    and row of the grid. (This is to allow irregular grid). It is also
    possible to use extra data in conjunction with other types of
    field, eg. a character string description of the field. The header
    variable LBEXT shows the length of extra data associated with a
    field. Thus the total length (in words) of the data record
    (LBLREC) equals the number of grid-point values plus LBEXT. Extra
    data is arranged in vectors, each of which is made up of an
    integer code followed by data values. The integer code is (1000xIA
    + IB), where IA is the length of the vector in words (i.e. usually
    the number of data values) and IB is a standard code. A zero
    integer code indicates no (more) extra data.  For cross-sections
    the extra data MUST start with an x-vector followed by a yvector;
    optionally other extra data may follow.

    :parameter m: A PP field's metatdata object
    :type m: `PPMetadata`
    '''
    if m.lhd[lbext] <= 0:
        # Skip the 4th block control word
        ppfile.seek(_word_size, os.SEEK_CUR) 
        m.extra = {}
        return m

    extra = {}

    domain_titles = []

    len = 0        
    while len < m.lhd[lbext]:

        intcode = numpy.fromfile(ppfile, dtype=m.file_int32, count=1)

        # Stop if we have run out of extra data
        if not (intcode or intcode[0]):
            break

        # Find ia = size of this extra data variable, 
        #      ib = type of this extra data variable
        ia, ib = divmod(intcode[0], 1000)
        
        if ib in (10, 11):
            # Read characters from the file
            title = numpy.fromfile(ppfile, dtype=m.file_str32, count=ia)
            title = ''.join(title)
            if ib == 10:
                extra[_extra_data_name[ib]] = numpy.array([title])
            else:
                domain_titles.append(title)
        else:
            # Don't read a number array, but store its locastion in a
            # data object
            parameters = {'_file'     : ppfile,
                          '_start'    : ppfile.tell(),
                          '_dtype'    : m.file_float32,
                          '_shape'    : (ia,),
                          }        
            extra[_extra_data_name[ib]] = Variable(**parameters)
            
            # Skip to the next extra data array
            ppfile.seek(ia*_word_size, os.SEEK_CUR)
            
        len += ia + 1
    #--- End: while    

    for bounds_type in ('', '_domain'):

        for dim in ('x', 'y'):
            dim_bounds_type = '%s%s' % (dim, bounds_type)
            lower  = '%s_lower_bound' % dim_bounds_type
            upper  = '%s_upper_bound' % dim_bounds_type
            bounds = '%s_bounds'      % dim_bounds_type
            
            try:
                extra[lower]
                extra[upper]
            except KeyError:
                pass
            else:
                extra[bounds] = VariableBounds(extra[lower], extra[upper])         
                extra.pop(lower)
                extra.pop(upper)
    #--- End: for

    # If we have domain titles then add them to the extra data
    # dictionary as a numpy array
    if domain_titles:
        extra[_extra_data_name[11]] = numpy.array(domain_titles)

    # Skip the 4th block control word so that we're ready to read the
    # next PP field
    ppfile.seek(_word_size, os.SEEK_CUR)            

    # Add the extra data to the metadata
    m.extra = extra

    # Return the updated metadata
    return m
#--- End: def

def _data(ppfile, m=None):
    
    '''
    :parameter ppfile: dch 

    :type ppfile: `file`

    :parameter m: A PP field's complete metatdata
    :type m: `PPMetadata`

    :returns: A `.Variable` object
    '''

    len_data = m.lhd[lblrec] - m.lhd[lbext]

    if len_data == 0:
        return None

    # Set the data type
    if m.lhd[lbuser1] == 1:
        datatype = m.file_float32
    elif m.lhd[lbuser1] == 2:
        datatype = m.file_int32
    elif m.lhd[lbuser1] == 3:
        # No idea if this works, or is valid code, or if I should even
        # consider it at all. Were logicals ever made part of PP? dch
        datatype = numpy.dtype('>b4')
    else:
        # Unknown data type => assume real
        datatype = m.file_float32

    # Create the data object
    parameters = {'_file'     : ppfile,
                  '_start'    : ppfile.tell(),
                  '_shape'    : (m.lhd[lbnpt], m.lhd[lbrow]),
                  '_dtype'    : datatype,
                  }    
    data = Variable(**parameters)

    # Skip data and 4th block control word to get to the start of the
    # next PP field in the file
    ppfile.seek(len_data*_word_size, os.SEEK_CUR)

    # Return the data object
    return data
#--- End: def

def _coordinate_data(coord, m=None,
                     data=None, bounds=None, climatology=None, 
                     coord_type=None, extra_type=None,
                     parameters={}):
    '''
    :parameter m: A PP field's complete metatdata
    :type m: `PPMetadata`
    '''
    if coord is None:
        return None

    create_bounds = True

    #-----------------------------------
    # Set time data from vtime and dtime
    #-----------------------------------
    if coord_type == 't':
        # This PP field's data array does not have time as one of
        # its two axes, so create a size 1 dimension coordinate for time
        data, bounds, climatology = None, None, None
        ib   = m.lbtim_ib
        if ib <= 1 or m.vtime >= m.dtime:
            data = [m.vtime]
        else:
            data = [0.5*(m.vtime + m.dtime)]
            array = [m.vtime, m.dtime]
            if ib == 2:
                bounds = array
            elif ib == 3:
                climatology = array

    # ------------------------------------------------
    # Set data from given values
    # ------------------------------------------------
    if data is not None:
        # Create a coordinate data object
        coord._data = Data(numpy.array(data))
    if bounds is not None:
        # Create a coordinate bounds data object
        coord.bounds       = cf.CoordinateBounds()
        coord.bounds._data = Data(numpy.array(bounds))
    elif climatology is not None:
        # Create a coordinate climatology data object
        coord.climatology       = cf.CoordinateBounds()
        coord.climatology._data = Data(numpy.array(climatology))

    # ------------------------------------------------

    # ------------------------------------------------
    if coord_type in ('x', 'y'):
        if _delta_zero(coord_type, m=m) or m.timeseries: # DCH Set a
            # Create a coordinate data object from the field's extra
            # data
            try:
                coord._data = Data(m.extra[coord_type])
            except KeyError: 
                pass

            # Create a coordinate bounds data object from the field's
            # extra data
            try:
                m.extra[coord_type+'_bounds']
            except KeyError: 
                pass
            else:
                coord.bounds       = cf.CoordinateBounds()
                coord.bounds._data = Data(m.extra[coord_type+'_bounds'])

        else:
            # Find, from the PP header, parameters for creating
            # coordinate data and coordinate bounds data objects
            if coord_type == 'x':
                parameters = {'_pseudo_origin': m.bhd[bzx],
                              '_delta'        : m.bhd[bdx],
                              '_shape'        : (m.lhd[lbnpt],),
                              '_dtype'        : m.float32,
                              }
                if m.ix in (13, 40, 99):
                    create_bounds = False                    
            else:
                parameters = {'_pseudo_origin': m.bhd[bzy],
                              '_delta'        : m.bhd[bdy],
                              '_shape'        : (m.lhd[lbrow],),
                              '_dtype'        : m.float32,
                              }        
                if m.iy in (13, 40, 99):
                    create_bounds = False

    if extra_type is not None:
        try:
            m.extra[extra_type] 
        except KeyError: 
            pass
        else:
            if not extra_type.endswith('bounds'):
                # Create a coordinate data object from the field's extra
                # data
                coord._data         = Data(m.extra[extra_type])
            else:
                # Create a coordinate bounds data object from the field's
                # extra data
                coord.bounds       = cf.CoordinateBounds()
                coord.bounds._data = Data(m.extra[extra_type])
    #--- End: if

    if parameters:
        # Create a coordinate data object
        coord._data = Data(VariableCalc(**parameters))

        # Create a coordinate bounds data object
        if create_bounds:
            parameters['_shape'] += (2,)            
            coord.bounds       = cf.CoordinateBounds()
            coord.bounds._data = Data(VariableCalcBounds(**parameters))
    #--- End: if 

    # Return coordinate
    return coord
#--- End: def

def _create_Coordinate(f, key, axis_code=-1, m=None, attr={},
                       data=None, bounds=None, climatology=None, 
                       coord_type=None, extra_type=None,
                       parameters=None, dimensions=[]):
    '''
    Create cf coordinate, giving it attributes and data

    :parameter m: A PP field's complete metatdata
    :type m: `PPMetadata`

    '''

    if axis_code is None:
        return None

    coord = cf.Coordinate()

    if axis_code >= 0:
        # Set attributes
        coord.axis          = _coord_axis[axis_code]
        coord.positive      = _coord_positive[axis_code]
        coord.standard_name = _coord_standard_name[axis_code]
        coord.units         = _coord_units[axis_code]   
        coord.ncvar         = coord.standard_name # dch look at cf.write and get rid of this
        
        if coord.standard_name is None:
            coord.long_name = _coord_long_name[axis_code]             
            del coord.standard_name
            coord.ncvar = coord.long_name # dch

        # Special cases
        if axis_code in (10, 11) and 100 < m.lhd[lbcode] < 10000:
            # Rotated lat-lon
            coord.units = 'degrees'        
        elif axis_code in (20, 23):
            # Time
            coord.units    = m.time_units
            coord.calendar = m.calendar
        
        # Remove None attributes        
        [delattr(coord, prop) 
         for prop in ('axis', 'positive') 
         if getattr(coord, prop) is None]
    #--- End: if

    # Apply extra attributes, or overwite existing ones
    for prop, value in attr.iteritems():
        if value is not None:
            setattr(coord, prop, value)
        else:
            try:
                delattr(coord, prop)
            except AttributeError: 
                pass

    # data
    coord = _coordinate_data(coord, m=m,
                             data        = data,
                             bounds      = bounds, 
                             climatology = climatology,
                             coord_type  = coord_type,
                             extra_type  = extra_type, 
                             parameters  = parameters)

    f.space.dimensions[key] = dimensions
    if key.startswith('d'):
        f.space.dimension_sizes[key] = _dimension_size(coord, coord_type=coord_type, m=m)        
        f.space.increasing[key]      = _dim_is_increasing(coord)

    coord = set_Data_attributes(f, key, coord)

    f.space[key] = coord 

    return coord
#--- End: def

def _dimension_size(coord, coord_type=None, m=None):
    '''
:parameter m: A PP field's complete metatdata
:type      m: `PPMetadata`
    '''
    try:
        size = coord.size
    except AttributeError:
        if coord_type == 'x':
            size = m.lhd[lbnpt]
        elif coord_type == 'y':
            size = m.lhd[lbrow]
        else:
            raise RuntimeError('Can\'t find dimension size for coord_type '+\
                                   coord_type)
    return size
#--- End: def

def _decode_exp(lbexp):
    '''
    
Decode the value of LBEXP in the PP header.

:parameter lbexp: The value of LBEXP in the PP header
:type lbexp: `int`

:parameter cache_runid: *Optional*.  
:type cache_runid: `dict`

:returns: A 5-character string derived from *lbexp*. If *lbexp* is
          negative then it is returned as a `string`.
'''
    try:
        # For speed, try to return a cached decoding of this lbexp
        return _cache['runid'][lbexp]
    except KeyError:
        pass

    if lbexp < 0:
        return str(lbexp)

    # Convert lbexp to a binary string, filled out to 30 bits with
    # zeros
    bits = bin(lbexp)
    bits = bits.lstrip('0b').zfill(30)

    # Step through 6 bits at a time, converting each 6 bit chunk into
    # a decimal integer, which is used as an index to the characters
    # lookup list.
    runid = []
    for i in xrange(0,30,6):
        index = int(bits[i:i+6], 2)
        runid.append(_characters[index])
    runid = ''.join(runid)

    # Enter this runid into the cache
    _cache['runid'][lbexp] = runid

    # Return the runid
    return runid
#--- End: def 

def _delta_zero(coord_type, m=None):
    '''
:parameter m: A PP field's complete metatdata
:type m: `PPMetadata`
    '''

    if coord_type == 'x':
        return abs(m.bhd[bdx]) <= cf.ATOL
    if coord_type == 'y':
        return abs(m.bhd[bdy]) <= cf.ATOL
#--- End: def

def _set_axis_codes(m=None):
    '''
    :parameter m: A PP field's complete metatdata
    :type m: `PPMetadata`
    '''
    m.ix, m.iy, m.iv, m.it = (None,) * 4

    code = m.lhd[lbcode]
    if code in (1, 101, 2, 102):
        m.ix, m.iy = 11, 10
    elif code >= 10000:
        temp,   ix = divmod(code, 10000)
        m.ix, m.iy = divmod(ix, 100)

    # Determine if we have a site-time cross section and, if so, if it
    # is a timeseries
    m.site_time_cross_section = False
    m.timeseries              = False
    if (m.ix, m.iy) == (13, 23):
        m.site_time_cross_section = True
        m.timeseries = m.lhd[lbuser3] == m.lhd[lbrow]

    # Set iv from LBVC
    try:
        m.iv = _lbvc_codes[m.lhd[lbvc]]
    except KeyError:
        pass

    # Set it
    if m.calendar == 'gregorian':
        m.it = 20
    else:
        m.it = 23

    return m
#--- End def

def _grid_dim_coordinate(f, coord, coord_type, m=None):
    '''
    s          : 
    coord      : 
    coord-type : 
    '''

    if coord is None:
        return s

    m.dimN += 1
    dim = 'dim'+str(m.dimN)
    m.dim = dim

    f.space.dimensions[dim]      = [dim]
    f.space.dimension_sizes[dim] = _dimension_size(coord, coord_type=coord_type, m=m)
    f.space[dim]                 = coord

    return f
#--- End: def

def _grid_aux_coordinate(f, coord, m=None, dimensions=[]):
    if coord is None:
        return f

    m.auxN += 1
    aux = 'aux'+str(m.auxN)   
    f.space.dimensions[aux] = dimensions
    f.space[aux]            = coord
    return f
#--- End: def

def _pp_header(ppfile):

    '''
:parameter ppfile: dch 
:type ppfile: `file`

:returns: A `PPMetadata` object containing all of the metadata for the
          current PP field in the input file, including some derived
          metadata quantities, such as 'calendar', 'time_units',
          'timeseries', etc.
'''

    m = PPMetadata()

    # Read 1st block control word
    bcw1 = numpy.fromfile(ppfile, dtype=m.file_int32, count=1)
    
    # Stop now if we have reached the end of the file
    if not bcw1:
        return None

    # Change data types if we have a 'wrong' endian PP file. ('Wrong'
    # simply means different to that assumed by class PPMetadata.)
#    if ppfile.tell() == 4: #m.npp == 1:
    if bcw1[0] == 65536:        
        m.file_float32 = m.float32.newbyteorder('S')
        m.file_int32   = m.int32.newbyteorder('S')
        m.file_str32   = m.str32.newbyteorder('S')
#        m.header_dtype = m.header_dtype.newbyteorder('S')
        
    elif bcw1[0] != 256:
        raise RuntimeError('Not a PP field: 1st block control word=%s' % bcw1[0])
#        d.datatype = numpy.dtype([('l', m.file_int32  , (45,)),
#                                  ('b', m.file_float32, (21,))
#                                  ])

    # Read the integer and real headers
#    m.hd = numpy.empty(1, m.header_dtype)

#    lhd = numpy.fromfile(ppfile, dtype=m.header_dtype, count=1)
    m.lhd = numpy.fromfile(ppfile, dtype=m.file_int32  , count=45)
    m.bhd = numpy.fromfile(ppfile, dtype=m.file_float32, count=21)

    # Create metadata object attributes out of header
#    for i in xrange(5000):
#        [setattr(m, _lhd_items[i], lhd[0][0][i]) for i in _xrange45]
#        [setattr(m, _bhd_items[i], lhd[0][1][i]) for i in _xrange19]
#    [setattr(m, _lhd_items[i], lhd[i]) for i in xrange(45)]
#    [setattr(m, _bhd_items[i], bhd[i]) for i in xrange(19)]
#    t1 = time.time()
#    _t['pp_h_io'] = t1-t0



#    # Create metadata object attributes out of header
#    for i in xrange(5000):
#        m.lhd = m.lhd
#        m.bhd[bhd = m.bhd[bhd
##        m.bhd[bhd = lhd[0][1]

    # Set some derived metadata quantities. Note that we need to this
    # before we get any extra data.
    m.lbtim_ia, ib         = divmod(m.lhd[lbtim], 100)
    m.lbtim_ib, m.lbtim_ic = divmod(ib, 10)

    m.time_units = 'days since 0-1-1'
    if m.lbtim_ic == 1:
        m.calendar = 'gregorian'
    else:
        m.calendar = '360_day'

    # For speed, try to return a cached numeric time value, otherwise
    # calculate one DCH ERROR here
    vtime = datetime.datetime(m.lhd[lbyr], m.lhd[lbmon], m.lhd[lbdat], 
                              m.lhd[lbhr], m.lhd[lbmin], 0)
    dtime = datetime.datetime(m.lhd[lbyrd], m.lhd[lbmond], m.lhd[lbdatd],
                              m.lhd[lbhrd], m.lhd[lbmind], 0)
    for attr, t in zip(('vtime', 'dtime'),
                       ( vtime ,  dtime)):
        try:
            value = _cache['time'][t]
        except KeyError:
            value = netCDF4.date2num(t, m.time_units, m.calendar)
            _cache['time'][t] = value
        setattr(m, attr, value)

    version, model = divmod(m.lhd[lbsrce], 10000)
    if model > 0:
        m.model = 'UM'
    if version > 0:
        m.version = str(version)
    
    m = _set_axis_codes(m=m)

#    m.file = os.path.basename(ppfile.name)


    # Return the metadata
    return m
#--- End: def

def _test_pp_condition(pp_condition, m=None):
    '''
    :parameter m: A PP field's complete metatdata
    :type m: `PPMetadata`
    '''
    if not pp_condition:
        return True

    elif pp_condition == 'true_latitude_longitude':
        if m.lhd[lbcode] in (1, 2):
            return True
        # Check pole location incase of incorrect LBCODE
        if (abs(m.bhd[bplat]-90.0) <= cf.ATOL + cf.RTOL*90.0 and 
            abs(m.bhd[bplon]) <= cf.ATOL):
            return True

    elif pp_condition == 'rotated_latitude_longitude':
        if m.lhd[lbcode] in (101, 102):
            return True
        # Check pole location incase of incorrect LBCODE
        if (abs(m.bhd[bplat]-90.0) <= cf.ATOL + cf.RTOL*90.0 and 
            abs(m.bhd[bplon]) <= cf.ATOL):
            return True

    else:
        raise RuntimeError("Unknown PP condition in STASH code conversion table: '"+\
                               pp_condition+"'")

    return False
#--- End: def


def _test_umversion(valid_from, valid_to, m=None):
    try:
        umversion = m.version
    except AttributeError:
        try:            
            umversion = _UMVERSION
        except NameError:
            raise RuntimeError('UM version is not set in PP header nor environment')

    if valid_from is '':
        valid_from = None

    if valid_to is '':
        if valid_from <= umversion:
            return True 
    elif valid_from <= umversion <= valid_to:
        return True 

    return False
#--- End: def


def _field_attributes(f, m=None):
    # Set CF attributes on the field

    f.set_att('pp'  , m)
    f.set_att('file', m.file)

    if abs(m.bhd[bdatum]) > cf.ATOL:
        f.add_offset = m.bhd[bdatum]

    f.Conventions = 'CF-1.5'
    
    f.runid = _decode_exp(m.lhd[lbexp])

    f._FillValue = m.bhd[bmdi]
    
    f.history = m.history

    if not abs(m.bhd[bmks]-1.0) <= cf.ATOL + cf.RTOL:
        f.scale_factor = m.bhd[bmks] # dch: check how this interacts with f.units        

    try:
        f.source = m.model
    except AttributeError:
        pass
    else:
        try:
            f.source = '%s vn%s' % (f.source, m.version)
        except AttributeError:
            pass        

    try:
        f.title = m.extra['title']
    except KeyError:
        pass

    # -------------------------------------------------------------
    # Set the field's long_name, standard_name and units from its
    # STASH code and possibly other properties
    # -------------------------------------------------------------
    stash = m.lhd[lbuser4]
    m.cf_info = {}
    if stash <= 0:
        # No STASH code set, record the STASH and field code
        # values in the long_name
        f.long_name = 'STASHcode=%s_Fieldcode=%s' % (stash, m.lhd[lbfc])
    else:

        # The STASH code has been set in the PP header, so try to find
        # its standard_name from the conversion table
        try:
            lines = _stash2standard_name[(m.lhd[lbuser7], stash)]
        except KeyError:
            pass
        else:
            for (long_name, 
                 units,
                 valid_from,
                 valid_to, 
                 standard_name,
                 cf_info,
                 pp_condition) in lines:
                
                # Check that conditions are met             
                version_ok      = _test_umversion(valid_from, valid_to, m=m)
                pp_condition_ok = _test_pp_condition(pp_condition, m=m)
                if not (version_ok and pp_condition_ok):                    
                    continue

                # Still here? Then we have our standard_name, etc.
                f.long_name = long_name.rstrip()
                f.units     = units
                if standard_name:
                    f.standard_name = standard_name
                    
                m.cf_info = cf_info
                    
                break
            #--- End: for            
        #--- End: try

        try:
            f.long_name
            try:   # dch this is a nightmare
                if not (f.units == '%' and f.scale_factor == 0.01):
                    f.units = '%s %s' % (f.scale_factor, f.units)
                    del f.scale_factor
            except AttributeError:
                pass
        except AttributeError:
            # Default long name in the absence of a conversion
            f.long_name = 'STASHcode=%s' % stash
    #--- End: if

    return f
#--- End: def

def read_stash2standard_name(delimiter='!'):
    ''' 

    Read the STASH to standard_name conversion table, as found at
    http://puma.nerc.ac.uk/STASH_to_CF/STASH_to_CF.txt, into a
    dictionary.

    :returns: A dictionary of tuples of lists

    *Examples*

    >>> stash2sn = read_stash2standard_name()
    >>> print stash2sn[(1,2)]
    (['U COMPNT OF WIND AFTER TIMESTEP     ',
      'm s-1',
      '',
      '',
      'eastward_wind',
      '',
      'true_latitude_longitude'],
     ['U COMPNT OF WIND AFTER TIMESTEP     ',
      'm s-1',
      '',
      '',
      'x_wind',
      '',
      'rotated_latitude_longitude'])
    '''
    lines = csv.reader(open('/home/david/STASH_to_CF.txt', 'r'), 
                       delimiter=delimiter, skipinitialspace=True)

    raw_list = []
    [raw_list.append(line) for line in lines]
        
    # Get rid of comments
    for line in raw_list[:]:
        if line[0].startswith('#'):
            raw_list.pop(0)
            continue
        break

    # Convert to a dictionary which is keyed by (submodel, STASHcode)
    # tuples
    stash2sn = {}
    for x in raw_list:
        key = (int(x[0]), int(x[1]))
        try:            
            cf_info = {}
            if x[7]:
                for d in x[7].split():
                    if d.startswith('height='): 
                        cf_info['height'] = re.split(_number_regex, d,
                                                     re.IGNORECASE)[1:4:2]
                        if cf_info['height'] == '':
                            cf_info['height'][1] = '1'

                    if d.startswith('below_'):
                        cf_info['below'] = re.split(_number_regex, d,
                                                     re.IGNORECASE)[1:4:2]
                        if cf_info['below'] == '':
                            cf_info['below'][1] = '1'

                    if d.startswith('where_'):         
                        cf_info['where'] = d.replace('where_', 'where ', 1)
                    if d.startswith('over_'):         
                        cf_info['over'] = d.replace('over_', 'over ', 1)

            x[7] = cf_info                    
        except IndexError:
            pass
        line = (x[2:],)

        try:
            stash2sn[key] += line
        except KeyError:
            stash2sn[key] = line

    return stash2sn
#--- End: def

# Create the STASH code to standard_name conversion dictionary at import time
_stash2standard_name = read_stash2standard_name()


def read(infile, exact=False, **kwargs):
    ''' 

    Read spaces from an input PP file on disk.
    
    The file may be big or little endian.

    All data arrays from all space components are not read from disk
    but are stored in the returned spaces as file pointers.


    :argument filename: An open file object, or a string giving the
                        file name, from which to read spaces.

    :type filename: `file` or `str`

       
    :parameter exact: *Optional*. If True then interpret the keywords
                      from `**kwargs` as exact phenomenon names. Refer
                      to `cf.Space.match` for details.

    :type exact: `bool`


    :parameter **kwargs: *Optional*. Other keyword arguments defining
                         space selection based on phenomena
                         criteria. If none are set then all spaces are
                         selected. Keywords and their arguments are
                         set as for `cf.Space.match`.


    :returns: A list of spaces (a `cf.FieldList` object).

    **Examples**
 
    >>> s = cf.pp.read('file.pp')

    '''

    # coord_type: A short string which acts as .... Valid values are
    # 't'  : time 
    #
    # 'v'  : an axis other than time which has the the same value for
    #        all data points, as described by LBVC
    #
    # 'p'  : pseudolevel
    #
    # 'rwl': radiation wavelength
    #
    # 'x'  : The X axis of the data array
    #
    # 'y'  : The Y axis of the data array

    # ------------------------------------------------------------
    # Preliminaries
    # ------------------------------------------------------------
    # Open the file, if it is not already open

    try:
        ppfile = open(infile, 'rb') 
    except TypeError:
        ppfile = infile

    # Initialize the output list of fields
    selected_fields_in_file = FieldList()

    

#    print 'dch ppfile.name =', ppfile.name

    history = '%s Converted from PP format by cf-python v%s' % \
        (time.strftime('%Y-%m-%d %H:%M:%S UTC', time.gmtime()),
         cf.__version__)

    filename = os.path.basename(ppfile.name)

    d = PPMetadata()

    dimN = -1
    auxN = -1

    # *************************************************************
    # Loop round each PP field in the file, reading it into a field
    # *************************************************************
    while True:

        # =========================================================
        # Get the next PP field's metadata and data and find some
        # derived metatdata quantities, such as 'calendar',
        # 'time_units', 'timeseries', etc.
        # =========================================================
        # Get the header (lhd and bhd)
        m = _pp_header(ppfile)

        # Stop now if we have reached the end of the file
        if m is None:
            break

        # Get the data
        data_array = _data(ppfile, m=m)

        # Get any extra data
        m = _extra_data(ppfile, m=m)

        m.history = history
        m.file    = filename

        # =========================================================
        # Initialize the field. The involves setting the field's
        # attributes and data and giving it an empty grid, for now.
        # =========================================================
        f       = Field()
        f.space = Space()
        
        # Set the field's attributes
        f = _field_attributes(f, m=m)

        axis_name = {}
        
        # ======================================================
        # Create the field's space
        # ======================================================
       
        # -----------------------------------------------------
        # Create a TIME dimension coordinate
        # -----------------------------------------------------
        dimN += 1         # Increment dimension number
        dim  = 'dim%(dimN)d' % locals()
        tdim = dim      

        # Set coordinate type and axis code
        if m.ix in (20,23):
            coord_type = 'x'
            axis_code  = m.ix
            # Flag that we have already created an 'X' coordinate
            m.ix = None
        elif m.iy in (20,23):
            coord_type = 'y'
            axis_code  = m.iy
            # Flag that we have already created a 'Y' coordinate
            m.iy = None
        else:
            coord_type = 't'
            axis_code  = m.it
            
        # Set the coordinate's data
        if m.timeseries:
            # This PP field is a timeseries. The validity time is
            # taken to be the time for the first sample, the data time
            # for the last sample, with the others evenly between.
            p = {'_shape' : (m.lhd[lbrow],),
                 '_dtype' : m.float32,}
            p['_delta'] = (m.dtime-m.vtime)/(p['_shape'][0]-1)
            p['_pseudo_origin'] = m.vtime - p['_delta']
        
            coord = _create_Coordinate(f, tdim, axis_code, m=m, parameters=p, 
                                       dimensions=[tdim])       
        else:
            coord = _create_Coordinate(f, tdim, axis_code, m=m, coord_type=coord_type, 
                                       dimensions=[tdim])
       
        # Save the dimension name
        if coord_type == 'x':
            data_dim0 = tdim
        elif coord_type == 'y' or m.timeseries:
            data_dim1 = tdim

        # Set the axis name for cell methods
        axis_name['t'] = coord.standard_name
        
        # ------------------------------------------------------
        # Create a 'VERTICAL' dimension coordinate. This size 1
        # coordinate will often by a vertical coordinate such as
        # pressure, but could be anything, such as longitude.
        # ------------------------------------------------------
        if m.iv is not None: 
            dimN += 1         # Increment dimension number
            dim  = 'dim%(dimN)d' % locals()
            vdim = dim      
     
            coord_type = 'v'
            try:
                height, units = m.cf_info['height']
            except (KeyError, TypeError):
                # Create the height coordinate from the PP header
                if m.lhd[lbvc] == 9:
                    pstar  = 1.0e5
                    data   = m.bhd[blev] + m.bhd[bhlev]/pstar
                    bounds = [m.bhd[brlev] +m.bhd[bhrlev]/pstar, 
                              m.bhd[brsvd1]+m.bhd[brsvd2]/pstar]
                elif m.lhd[lbvc] == 65:
                    pass
                else:
                    data   = [m.bhd[blev]],
                    bounds = [m.bhd[brlev], m.bhd[brsvd1]]                
                coord = _create_Coordinate(f, vdim, m.iv, m=m, data=data, bounds=bounds,
                                           dimensions=[vdim])    
            else:               
                # Create the height coordinate from the information
                # given in the STASH to standard_name conversion table
                coord = _create_Coordinate(f, 15, m=m, attr={'units': units},
                                           data = [float(height)],
                                           dimensions=[vdim])
            #--- End: try

            # Set the axis name for cell methods
            axis_name[coord_type] = coord.standard_name
        #--- End: if

        # ------------------------------------------------------
        # Create a PSEUDOLEVEL dimension coordinate
        # ------------------------------------------------------
        if m.lhd[lbuser5] > 0: 

            dimN += 1         # Increment dimension number
            dim  = 'dim%(dimN)d' % locals()
            pdim = dim      
            
            coord = _create_Coordinate(f, pdim, 40, m=m, data=[m.lhd[lbuser5]],
                                       dimensions=[pdim])
        
            # Set the axis name for cell methods (as yet there is no
            # standard_name for this, so use the long_name)
            axis_name['p'] = coord.long_name
        #--- End: if

        # --------------------------------------------------------
        # Create a dimension coordinate for 'RADIATION WAVELENGTH'
        # --------------------------------------------------------
            #dch : do this instead of VERTICAL, I wonder
        try:
            rwl, units = m.cf_info['below']
        except (KeyError, TypeError):
            pass
        else:
            
            dimN += 1         # Increment dimension number
            dim    = 'dim%(dimN)d' % locals()
            rwldim = dim      
            
            # Create the radtion wavelength coordinate
            attr = {'standard_name' : 'radiation_wavelength', 
                    'units'         : units}
            coord = _create_Coordinate(f, rwldim, m=m, attr=attr,
                                       data=[rwl], bounds=[0.0, rwl],
                                       dimensions=[rwldim])
            
            # Set the axis name for cell methods
            axis_name['rwl'] = coord.standard_name
        #--- End: try

        # ----------------------------------------------------------
        # Create 'X' and 'Y' dimension coordinates, if they haven't
        # already been created. These coordinates represent the 'x'
        # and 'y' dimensions of the data array. In the case of cross
        # sections, one of them may have already been created.
        # ----------------------------------------------------------
        # Set coordinate type and axis code
        for coord_type, axis_code in zip(('x' , 'y' ),
                                         (m.ix, m.iy)):
    
            dimN += 1         # Increment dimension number
            dim   = 'dim%(dimN)d' % locals()           
       
            # Skip this coordinate if it has already been created
            if axis_code is None:
                continue    

            coord = _create_Coordinate(f, dim, axis_code, m=m,
                                       coord_type=coord_type,
                                       dimensions=[dim])
        
            if coord_type == 'x':
                xdim      = dim
                data_dim0 = xdim
            else:
                ydim      = dim
                data_dim1 = ydim

            # Set the axis name for cell methods
            axis_name[coord_type] = coord.standard_name
        #--- End: for

        # -------------------------------------------------------------------------
        # Create size 1, 1d auxiliaries
        # -------------------------------------------------------------------------
        t0=t1
        # Forecast reference time
        if m.lbtim_ib == 1: 
            auxN += 1         # Increment auxiliary number
            aux   = 'aux%(auxN)d' % locals()           
       
            attr = {'standard_name': 'forecast_reference_time',
                    'units'        : m.time_units,
                    }
            coord = _create_Coordinate(f, aux, m=m, attr=attr, data=[m.dtime],
                                       dimensions=[tdim])           
        #--- End: if

        # -------------------------------------------------------
        # Create a model level number auxiliary coordinate
        # -------------------------------------------------------
        lev = m.lhd[lblev]
        if (m.iv is not None and 
            auxN += 1         # Increment auxiliary number
            aux   = 'aux%(auxN)d' % locals()           

            m.lhd[lbvc] in (9, 65) and lev not in (7777, 8888, 9999)): # dch check 7777

            coord = _create_Coordinate(f, aux, 5, m=m, data=[lev], dimensions=[vdim])
        #--- End: if

        # -------------------------------------------------------
        # Create domain title auxiliary coordinate
        # -------------------------------------------------------
        try:            
            domain_titles = m.extra['domain_title']
        except KeyError:
            pass
        else:
            dim = None
            if ix == 13:
                dim = xdim
            elif iy == 13:
                dim = ydim
            if dim:
                auxN += 1                        # Increment auxiliary number
                aux   = 'aux%(auxN)d' % locals()           

                attr = {'units'    : 'string',
                        'long_name': 'domain title'}
                coord = _create_Coordinate(f, aux, m=m, data=domain_titles, attr=attr,
                                           dimensions=[dim])
        #--- End: try

        # -------------------------------------------------------
        # Create two dimensional unrotated latitude and longitude
        # auxiliary coordinates.
        # -------------------------------------------------------
        if f.space[data_dim0].units == f.space[data_dim1].units == 'degrees':

            grid_mapping_name = 'rotated_latitude_longitude'

            # Try to find the appropiate unrotated latitude and
            # longitude arrays in the cache
            found_cached_data = None
            
            for rot in _cache['latlon']:
                if (rot[0]._data.equals(f.space[data_dim0]._data) and 
                    rot[1]._data.equals(f.space[data_dim1]._data)):
                    latlon = _cache['latlon'][rot] 
                    if f.space[data_dim0].standard_name == 'longitude':
                        dimensions = [data_dim1, data_dim0]                        
                    else:
                        dimensions = [data_dim0, data_dim1]
                    found_cached_data = True
                    break
            #--- End: for

            # Make the unrotated latitude and longitude arrays if we
            # couldn't find them in the cache
            if not found_cached_data:
                print 'dch CALC'
                if f.space[data_dim0].standard_name == 'longitude':
                    dimensions = [data_dim1, data_dim0]
                    rotated_lat = f.space[data_dim1]
                    rotated_lon = f.space[data_dim0]
                else:
                    dimensions = [data_dim0, data_dim1]
                    rotated_lat = f.space[data_dim0]
                    rotated_lon = f.space[data_dim1]
                latlon = 0

                latlon = unrotated_latlon(rotated_lat.varray, rotated_lon.varray,
                                          m.bhd[bplat], m.bhd[bplon])                
                
                f.space[data_dim0].repoint()          
                f.space[data_dim1].repoint()

                # Add the new lat lon to the cache
                _cache['latlon'][(f.space[data_dim0], f.space[data_dim1])] = latlon
            #--- End: if
             
            # Create the unrotated latitude and longitude coordinates        
#            axis_code         = (10, 11)
#            units             = ('degrees_north', 'degrees_east')
#            standard_name     = ('latitude'     , 'longitude')
            
            
            for standard_name, units in zip(('latitude'     , 'longitude'   ),
                                            ('degrees_north', 'degrees_east')):
                auxN += 1                        # Increment auxiliary number
                aux   = 'aux%(auxN)d' % locals()           

                attr = {'standard_name': standard_name,
                        'units'        : units,
                        'transform'    : grid_mapping_name,
                        }                
#dch does this need an axis code??
                coord = _create_Coordinate(f, aux, m=m, attr=attr, data=latlon[i,...],
                                           dimensions=dimensions)
            #--- End: for

            # Create the tranform construct
            transform = cf.Transform()
            transform.transform_name = grid_mapping_name
            transform['grid_mapping_name'] = grid_mapping_name
            transform['grid_north_pole_latitude']  = m.bhd[bplat]
            transform['grid_north_pole_longitude'] = m.bhd[bplon]
            try:
                f.space.transform
            except AttributeError:
                f.space.transform = cf.CfDict()     
            f.space.transform[grid_mapping_name] = transform
        #--- End: if

        # ------------------------------------------------------------
        # Create auxiliary coordinates for the latitudes and
        # longitudes of the sites in site-time cross sections. (Don't
        # consider 'z' extra data because we don't know what it refers
        # to.)
        # ------------------------------------------------------------
        if m.site_time_cross_section:

            for axis_code, extra_type in zip((11 , 10 ),

                try:
                    coord_type = '%s_domain_bounds' % extra_type
                    m.extra[coord_type]
                except KeyError:
                    try:
                        coord_type = '%s_domain_lower_bound' % extra_type
                        m.extra[coord_type]
                    except KeyError:
                        pass
                    else:           
                        # Create, from extra data, an auxiliary
                        # coordinate with data but no bounds, if the
                        # data noes not contain any missing values
                        file_position = ppfile.tell()
                        data = m.extra[coord_type][...]
                        # Reset the file pointer after reading the
                        # extra data into a numpy array
                        ppfile.seek(file_position, os.SEEK_SET)
                        if not numpy.any(data == _pp_rmdi): # dch also test in bmdi?   
                            auxN += 1                        # Increment auxiliary number
                            aux   = 'aux%(auxN)d' % locals()           
                            coord = _create_Coordinate(f, aux, axis_code, m=m, data=data,
                                                       attr={'axis': None}, dimensions=[xdim])# DCH      
                else:                                                                         # xdim?    
                    # Create, from extra data, an auxiliary coordinate                        # should   
                    # with 1) data and bounds, if the upper and lower                         # be       
                    # bounds have no missing values; or 2) data but no                        # the      
                    # bounds, if the upper bound has missing values                           # axis     
                    # but the lower bound does not.                                           # which    
                    file_position = ppfile.tell()                                             # has      
                    bounds = m.extra[coord_type][...]                                         # axis_code
                    # Reset the file pointer after reading the extra                          # 13       
                    # data into a numpy array
                    ppfile.seek(file_position, os.SEEK_SET)
                    data = None
                    if numpy.any(bounds[...,1] == _pp_rmdi): # dch also test in bmdi?
                        if not numpy.any(bounds[...,0] == _pp_rmdi): # dch also test in bmdi?
                            data = bounds[...,0]
                        bounds = None
                    else:
                        data = numpy.mean(bounds, axis=1)
                    if (data, bounds) != (None, None):
                        auxN += 1                        # Increment auxiliary number
                        aux   = 'aux%(auxN)d' % locals()           
                        
                        coord = _create_Coordinate(f, aux, axis_code, m=m,
                                                   data       = data,
                                                   bounds     = bounds, 
                                                   attr       = {'axis': None},
                                                   dimensions = [xdim]) # DCH      
                                                                        # xdim?    
                                                                        # should   
                                                                        # be       
                                                                        # the      
                                                                        # axis     
                                                                        # which    
                                                                        # has      
                                                                        # axis_code
                                                                        # 13       
               #--- End: try
           #--- End: for
       #--- End: if
        
        # ---------------------------------------------------------
        # Set field' data dimension names
        # ---------------------------------------------------------
        f.space.dimensions['data'] = [data_dim0, data_dim1]

        # ---------------------------------------------------------
        # Set the field's data and the data's attributes
        # ---------------------------------------------------------
        if data_array is not None:
            f._data = Data(data_array)          
            f       = set_Data_attributes(f, 'data', f)

        # =========================================================
        # Create the field's cell methods
        # =========================================================
        cell_methods = []
        f.space.nc_dimensions = {}

        proc       = m.lhd[lbproc]       
        tmean_proc = 0
        if m.lbtim_ib in (2, 3):
            if proc-128 in (0, 64, 2048, 4096, 8192):
                tmean_proc = 128
                proc -= 128
                
        # ---------------------------------------------------------
        # Area cell methods
        # ---------------------------------------------------------
        if m.ix in (10, 11, 12) and m.iy in (10, 11, 12):
            if proc == 64 or m.cf_info.has_key('where'):
                cell_methods.append('area: mean')

            try:
                cell_methods.append(m.cf_info['where'])
            except (KeyError, TypeError):
                pass
            else:
                try:
                    cell_methods.append(m.cf_info['over'])
                except (KeyError, TypeError):
                    pass
#            for info in m.cf_info:                
#                if info.startswith('area: mean where_'):                    
#                    cell_methods.append(info.replace('where_', 'where ', 1))
#                elif info.startswith('over_'):
#                    cell_methods.append(info.replace('over_', 'over ', 1))
        # dch : do speial zonal mean as as in pp_cfwrite
        
        # ---------------------------------------------------------
        # Vertical cell methods
        # ---------------------------------------------------------
        if proc == 2048:
            v_axisname = axis_name['v']
            cell_methods.append('%s: mean' % v_axisname)
            f.space.nc_dimensions[vdim] = v_axisname

        # ---------------------------------------------------------
        # Time cell methods
        # ---------------------------------------------------------
        t_axisname = axis_name['t']
        f.space.nc_dimensions[tdim] = axis_name['t']
        if m.lbtim_ib == 0:
            cell_methods.append('%s: point' % t_axisname)
        elif proc == 4096:
            cell_methods.append('%s: minimum' % t_axisname)
        elif proc == 8192:
            cell_methods.append('%s: maximum' % t_axisname)
        if tmean_proc == 128:
            if m.lbtim_ib == 2:
                cell_methods.append('%s: mean' % t_axisname)
            elif m.lbtim_ib == 3:
                cell_methods.append('%s: mean within years' % t_axisname)
                cell_methods.append('%s: mean over years' % t_axisname) 

#        print m.lhd[lbproc], proc, tmean_proc, cell_methods, m.lbtim_ib
        # ---------------------------------------------------------
        # Add the cell methods to the field
        # ---------------------------------------------------------
        if cell_methods:
            f.cell_methods = cf.CellMethods()
            f.cell_methods = f.cell_methods.parse(' '.join(cell_methods), 
                                                  field=f)
            

        # ============================================================
        # The field is complete(!). Add it to the list of fields to be
        # returned if it matches any given phenomena criteria
        # ============================================================
        if f.match(exact=exact, **kwargs):            
            f.squash()
            selected_fields_in_file.append(f)


        # ============================================================
        # Now read the next PP field in the file
        # ============================================================
    # --- End: while

    # Close the ppfile
    ppfile.close()

    # Return the selected list of fields
    return selected_fields_in_file
#--- End: def
