import cf
import tempfile
import os
import sys
import numpy

'''
Test the cf package:

TEST: Create a space in memory
TEST: Print a dump of the space
TEST: Write the space to disk
TEST: Read the space from disk, with data stored as netCDF4.Variable objects
TEST: Access the space's data as a numpy array
TEST: Access the space's coordinates' data as a numpy arrays
TEST: Slice the space
TEST: Check the equality function
TEST: Create list of spaces
TEST: Write a list of spaces to disk
TEST: Read a list of spaces from disk
TEST: Print all last spaces in the list
TEST: Print the last space in the list
TEST: Print the data of the last space in the list
TEST: Modify the last space in the list
'''

print '\nTEST: Create a space in memory:'

s = cf.Space()
s.grid = cf.Grid()

s.grid['dim0'] = cf.Coordinate()
s.grid['dim0'].standard_name = 'latitude'
s.grid['dim0']._data = cf.Data(numpy.arange(10, dtype = float))

s.grid['dim1'] = cf.Coordinate()
s.grid['dim1'].standard_name = 'longitude'
s.grid['dim1']._data = cf.Data(numpy.arange(9, dtype = float) + 20)

s.grid.dimensions = {'data': ['dim0', 'dim1'],
                     'dim0': ['dim0'],
                     'dim1': ['dim1'],
                     }

s.grid.dimension_sizes = {'dim0': 10,
                          'dim1': 9,
                          }

s.standard_name = 'test_space'
s._data = cf.Data(numpy.arange(90, dtype = float).reshape(10,9))

print '\nTEST: Print a dump of the space:'
print repr(s)

cf.dump(s)

print '\nTEST: Write the space to disk:'
file = tempfile.mktemp('.nc')

cf.write(s, file)

print '\nTEST: Read the space from disk, with data stored as netCDF4.Variable objects: '
t = cf.read(file)[0]

cf.dump(t)

print '\nTEST: Access the space\'s data as a numpy array:'
print t.varray

print '\nTEST: Access the space\'s coordinates\' data as a numpy arrays:'

print t.coord('lat').varray
print t.coord('lon').varray

print '\nTEST: Slice the space:'
print t.slice(lat = cf.lt(5), lon = 27)

print '\nTEST: Slice the space:'
print t.slice[...,2:5]

print '\nTEST: Check the equality function:'
result  = cf.equals(t, t)

if not result:
    raise RuntimeError('Space is not equal to itself')

print '\nTEST: Create list of spaces:'
sl = cf.SpaceList([t,t,t,t])

print '\nTEST: Write a list of spaces to disk:'
cf.write(sl, file)

print '\nTEST: Read a list of spaces from disk:'
sl = cf.read(file)
print repr(sl)

print '\nTEST: Print all last spaces in the list:'
print sl

print '\nTEST: Print the last space in the list:'
print sl[-1]

print '\nTEST: Print the data of the last space in the list:'
print sl[-1].varray

print '\nTEST: Modify the last space in the list:'
sl[-1] *= -1
print sl[-1].varray

print '\n-------------------------------------------'
print   'TEST: All tests passed for cf version',cf.__version__
print   '------------------------------------------- '

# Tidy up
os.remove(file)

